
static const u32 RingBuffer_Size=8*1024; // req power of 2
static const u32 RingBuffer_DstSampleRate=44100;
typedef struct {
  u32 SrcSampleRate,SamplesPerFlame;
  volatile u32 ReadIndex,WriteIndex;
  u32 *pbuf;
  u32 freqcur,freqadd;
  s16 freqL0,freqR0,freqL1,freqR1;
  u32 *pFlameBuf;
} TRingBuffer;

static TRingBuffer RingBuffer;

static void audioRingBufferCallback(void* buf, unsigned int length, void *userdata);

static volatile bool audioRingBufferInterrupt;

static void audioRingBufferInit(void)
{
  TRingBuffer *prb=&RingBuffer;
  
  prb->SrcSampleRate=0;
  prb->SamplesPerFlame=0;
  
  prb->ReadIndex=0;
  prb->WriteIndex=0;
  
  prb->pbuf=(u32*)malloc(RingBuffer_Size*4);
  for(u32 idx=0;idx<RingBuffer_Size;idx++){
    prb->pbuf[idx]=0;
  }
  
  prb->pFlameBuf=NULL;
  
  prb->freqcur=0;
  prb->freqadd=0;
  
  prb->freqL0=0;
  prb->freqR0=0;
  prb->freqL1=0;
  prb->freqR1=0;
  
  audioRingBufferInterrupt=false;
  
  pspAudioSetVolume(0, PSP_VOLUME_MAX,PSP_VOLUME_MAX);
  pspAudioSetChannelCallback(0, audioRingBufferCallback, NULL);
}

static void audioRingBufferSetVolume15(u32 Volume15)
{
  pspAudioSetVolume(0, Volume15,Volume15);
}

static void MT_Start(void);
static void MT_End(void);
static void MT_RequestUpdate(void);
static void MT_RequestSeekSec(float sec);

static void audioRingBufferStart(u32 SrcSampleRate,u32 SamplesPerFlame)
{
  TRingBuffer *prb=&RingBuffer;
  
  if(prb->SrcSampleRate!=SrcSampleRate){
    while(1){
      u32 ridx=prb->ReadIndex;
      const u32 widx=prb->WriteIndex;
      u32 last=(RingBuffer_Size+widx-ridx)&(RingBuffer_Size-1);
      if(last==0) break;
      sceDisplayWaitVblankStart();
    }
    while(1){
      if(audioRingBufferInterrupt==false) break;
      sceDisplayWaitVblankStart();
    }
  }
  
  if((RingBuffer_Size-2048)<SamplesPerFlame){
    printf("Flame size overflow. (%d-2048)<%d\n",RingBuffer_Size,SamplesPerFlame);
    SystemHalt();
  }
  
  prb->SrcSampleRate=SrcSampleRate;
  prb->SamplesPerFlame=SamplesPerFlame;
  
  prb->pFlameBuf=(u32*)malloc(prb->SamplesPerFlame*4);
  for(u32 idx=0;idx<prb->SamplesPerFlame;idx++){
    prb->pFlameBuf[idx]=0;
  }
  
  prb->freqcur=0;
  prb->freqadd=prb->SrcSampleRate*0x10000/RingBuffer_DstSampleRate;
  
  prb->freqL0=0;
  prb->freqR0=0;
  prb->freqL1=0;
  prb->freqR1=0;
  
  MT_Start();
}

static void audioRingBufferEnd(void)
{
  TRingBuffer *prb=&RingBuffer;
  
  MT_End();
  
  if(prb->pFlameBuf!=NULL){
    free(prb->pFlameBuf); prb->pFlameBuf=NULL;
  }
}

static void audioRingBufferFree(void)
{
  TRingBuffer *prb=&RingBuffer;
  
  pspAudioSetChannelCallback(0, NULL, NULL);
  
  while(1){
    if(audioRingBufferInterrupt==false) break;
    sceDisplayWaitVblankStart();
  }
  
  if(prb->pbuf!=NULL){
    free(prb->pbuf); prb->pbuf=NULL;
  }
}

static void audioRingBufferRequestSeekSec(float sec)
{
  MT_RequestSeekSec(sec);
}

static void audioRingBufferCallback(void* buf, unsigned int length, void *userdata)
{
  TRingBuffer *prb=&RingBuffer;
  
  audioRingBufferInterrupt=true;
  
  const u32 buflength=length;
  u32 *pbuf32=(u32*)buf;
  
  if(prb->pbuf==NULL){
    length=0;
    }else{
    u32 ridx=prb->ReadIndex;
    const u32 widx=prb->WriteIndex;
    
    u32 last=(RingBuffer_Size+widx-ridx)&(RingBuffer_Size-1);
    if(last<length) length=last;
    
    u32 cur=prb->freqcur;
    const u32 add=prb->freqadd;
    s16 L0=prb->freqL0;
    s16 R0=prb->freqR0;
    s16 L1=prb->freqL1;
    s16 R1=prb->freqR1;
    
    if(add==0x10000){
      for(u32 idx=0;idx<length;idx++){
        u32 smp=prb->pbuf[ridx];
        ridx=(ridx+1)&(RingBuffer_Size-1);
        pbuf32[idx]=smp;
      }
      }else{
      for(u32 idx=0;idx<length;idx++){
        cur+=add;
        while(0x10000<cur){
          cur-=0x10000;
          u32 smp=prb->pbuf[ridx];
          prb->pbuf[ridx]=0;
          ridx=(ridx+1)&(RingBuffer_Size-1);
          L0=L1; R0=R1;
          L1=smp&0xffff; R1=smp>>16;
        }
        s16 L=((L0*(0x10000-cur))+(L1*cur))/0x10000;
        s16 R=((R0*(0x10000-cur))+(R1*cur))/0x10000;
        pbuf32[idx]=(L&0xffff)|(R<<16);
      }
    }
    
    prb->freqcur=cur;
    prb->freqL0=L0;
    prb->freqR0=R0;
    prb->freqL1=L1;
    prb->freqR1=R1;
    
    prb->ReadIndex=ridx;
  }
  
  for(u32 idx=length;idx<buflength;idx++){
    pbuf32[idx]=0;
  }
  
  MT_RequestUpdate();
  
  audioRingBufferInterrupt=false;
}

// ------------------------------------------------------------------------

static bool MT_RequestExit;
static float MT_SeekSec;
static int MT_SemaID=-1;
static SceUID MT_ThreadID=-1;

static int MT_DecodeThread(SceSize args, void *argp)
{
  TRingBuffer *prb=&RingBuffer;
  
  while(1){
    if(MT_RequestExit==true) break;
    
    if(MT_SeekSec!=-1){
      pInterface->Seek(MT_SeekSec);
      MT_SeekSec=-1;
    }
    
    { // Check ring buffer.
      const u32 ridx=prb->ReadIndex;
      const u32 widx=prb->WriteIndex;
      u32 req=(RingBuffer_Size+ridx-widx)&(RingBuffer_Size-1);
      if(req==0) req=RingBuffer_Size;
      if(req<=prb->SamplesPerFlame){
        sceKernelWaitSema(MT_SemaID,1,0);
        sceKernelSignalSema(MT_SemaID,0);
        continue;
      }
    }
    
    u32 *pBufLR=prb->pFlameBuf;
    
    u32 SamplesCount=pInterface->Update(pBufLR);
    if(prb->SamplesPerFlame<SamplesCount){
      printf("Flame buffer overflow. %d<%d.\n",prb->SamplesPerFlame,SamplesCount);
    }
    if(SamplesCount==0) break;
    
    const u32 ridx=prb->ReadIndex;
    u32 widx=prb->WriteIndex;
    
    for(u32 idx=0;idx<SamplesCount;idx++){
      prb->pbuf[widx]=pBufLR[idx];
      widx=(widx+1)&(RingBuffer_Size-1);
    }
    
    prb->WriteIndex=widx;
  }
  
  return(0);
}

static void MT_Start(void)
{
  MT_RequestExit=false;
  MT_SeekSec=-1;
  
  MT_SemaID=sceKernelCreateSema("LibSnd_Decode_Sema",0,0,1,0);
  
  const char *pThreadName="LibSnd_DecodeThread";
  MT_ThreadID=sceKernelCreateThread(pThreadName,MT_DecodeThread,ThreadPrioLevel_LibSnd_Decode,0x10000,PSP_THREAD_ATTR_VFPU|PSP_THREAD_ATTR_USER|PSP_THREAD_ATTR_CLEAR_STACK,NULL);
  if(MT_ThreadID<0){
    printf("Error: Can not create thread. (ec:%d) [%s]\n",MT_ThreadID,pThreadName);
    SystemHalt();
  }
  
  int ret=sceKernelStartThread(MT_ThreadID,0,NULL);
  if(ret<0){
    printf("Error: Can not start thread. (ec:%d) [%s]\n",ret,pThreadName);
    SystemHalt();
  }
}

static void MT_End(void)
{
  if(MT_ThreadID!=-1){
    MT_RequestExit=true;
    MT_RequestUpdate();
    sceKernelWaitThreadEnd(MT_ThreadID,NULL);
    sceKernelDeleteThread(MT_ThreadID);
    MT_ThreadID=-1;
  }
  
  if(MT_SemaID!=-1){
    sceKernelDeleteSema(MT_SemaID);
    MT_SemaID=-1;
  }
  
  MT_SeekSec=-1;
}

static void MT_RequestUpdate(void)
{
  sceKernelSignalSema(MT_SemaID,1);
}

static void MT_RequestSeekSec(float sec)
{
  MT_SeekSec=sec;
}

