
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#include <pspuser.h>
#include <pspdisplay.h>
#include <pspgu.h>
#include <pspgum.h>
#include <psppower.h>
#include <pspctrl.h>

PSP_MODULE_INFO("MSB", PSP_MODULE_USER, 1, 1);
PSP_MAIN_THREAD_ATTR(THREAD_ATTR_USER | THREAD_ATTR_VFPU | PSP_THREAD_ATTR_CLEAR_STACK);

PSP_HEAP_SIZE_KB(20*1024);          // LR
PSP_MAIN_THREAD_STACK_SIZE_KB(512); // LR

#include "common.h"

#include "LibSnd.h"
#include "LibImg.h"

#include "BMPReader.h"
#include "memtools.h"
#include "strtool.h"
#include "CFont.h"
#include "euc2unicode.h"
#include "GU.h"
#include "VRAMManager.h"
#include "TexFont.h"
#include "sceHelper.h"
#include "PlayList.h"

//--------------------------------------------------------------------

#include "main_CallBacks_System.h"
#include "main_CPU.h"
#include "main_Keys.h"

//--------------------------------------------------------------------

void SystemHalt(void)
{
  sceKernelDcacheWritebackAll();
  while(1){
    sceDisplayWaitVblankStart();
    if(SystemExitRequest==true) break;
  }
  sceKernelExitGame();
}

static void PrintFreeMem(void)
{
  const u32 maxsize=24*1024*1024;
  const u32 segsize=1*1024;
  const u32 count=maxsize/segsize;
  u32 *pptrs=(u32*)malloc(count*4);
  
  if(pptrs==NULL){
    printf("PrintFreeMem: Investigation was interrupted. Very low free area.\n");
    SystemHalt();
  }
  
  u32 FreeMemSize=0;
  u32 MaxBlockSize=0;
  
  for(u32 idx=0;idx<count;idx++){
    u32 size=maxsize-(segsize*idx);
    pptrs[idx]=(u32)malloc(size);
    if(pptrs[idx]!=0){
      FreeMemSize+=size;
      if(MaxBlockSize<size) MaxBlockSize=size;
    }
  }
  
  printf("FreeMem=%dkbyte (MaxBlockSize=%dkbyte)\n",FreeMemSize/1024,MaxBlockSize/1024);
  
  for(u32 idx=0;idx<count;idx++){
    if(pptrs[idx]!=0){
      free((void*)pptrs[idx]); pptrs[idx]=0;
    }
  }
  
  if(pptrs!=NULL){
    free(pptrs); pptrs=NULL;
  }
}

//--------------------------------------------------------------------

static TProc_Interface *pProc_Interface;
EProcState ProcStateNext;

#include "SystemSmallTexFont.h"

TTexFont SystemSmallTexFont;

static void main_Start(void)
{
  SetupCallbacks();
  sceHelper_Init();
  
  VRAMManager_Init();
  VRAMManager_SytemMode_Start();
  
  GuOpen();
  
  Keys_Init();
  
  EUC2Unicode_Init();
  EUC2Unicode_Load();
  
  CFont_Init(ECFS_12);
  CFont_Init(ECFS_14);
  CFont_Init(ECFS_16);
  CFont_Init(ECFS_20);
  CFont_Init(ECFS_24);
  
  LibSnd_Init();
  LibImg_Init();
  
  PlayList_Init();
  
  TexFont_Create(&SystemSmallTexFont,EVMM_System,"!\"#$%&'()*+,-./0123456789:;<=>?@ABCDEFGHIJKLMNOPQRSTUVWXYZ[\\]^_|",Resources_FontPath "/SmallFont.png",ETF_RGBA8888);
}

static void main_End(void)
{
  PrintFreeMem();
  VRAMManager_ShowFreeMem();
  
  LibSnd_Close();
  LibSnd_Free();
  LibImg_Close();
  LibImg_Free();
  
  PlayList_Free();
  
  TexFont_Free(&SystemSmallTexFont,EVMM_System);
  
  EUC2Unicode_Free();
  
  CFont_Free(ECFS_12);
  CFont_Free(ECFS_14);
  CFont_Free(ECFS_16);
  
  Keys_Free();
  
  GuClose();
  
  PrintFreeMem();
  VRAMManager_ShowFreeMem();
  
  VRAMManager_SytemMode_End();
  VRAMManager_End();
  
  PrintFreeMem();
  VRAMManager_ShowFreeMem();
  
  sceHelper_Free();
}

static void main_ProcessKeys(u32 VSyncCount)
{
  TKeys *pk=Keys_Refresh();
  
  pProc_Interface->KeysUpdate(pk,VSyncCount);
  
  u32 KeyCursor=KeyRepeat_Update(&KeyRepeat_Cursor,pk->Buttons);
  u32 KeyButton=KeyRepeat_Update(&KeyRepeat_Button,pk->Buttons);
  u32 KeyTriggers=KeyRepeat_Update(&KeyRepeat_Triggers,pk->Buttons);
  
  u32 keys=KeyCursor|KeyButton|KeyTriggers;
  pProc_Interface->KeyPress(keys,VSyncCount);
}

static void main_Loop(void)
{
  ProcStateNext=EPS_FileList;
  
//  ProcStateNext=EPS_ImageView; snprintf(Proc_ImageView_ImageFilename,256,"/test.bmp");
  
  EProcState ProcStateLast=EPS_Loop;
  
  while(1){
    VRAMManager_ProcessMode_Start();
    
    KeyRepeat_Init();
    
    switch(ProcStateNext){
      case EPS_Loop: {
        printf("Illigal proc state. ProcStateNext:%d ProcStateLast:%d\n",ProcStateNext,ProcStateLast);
        SystemHalt();
      } break;
      case EPS_FileList: {
        extern TProc_Interface* Proc_FileList_GetInterface(void);
        pProc_Interface=Proc_FileList_GetInterface();
      } break;
      case EPS_ImageView: {
        extern TProc_Interface* Proc_ImageView_GetInterface(void);
        pProc_Interface=Proc_ImageView_GetInterface();
      } break;
      case EPS_TextReader: {
        extern TProc_Interface* Proc_TextReader_GetInterface(void);
        pProc_Interface=Proc_TextReader_GetInterface();
      } break;
      case EPS_MusicPlayer: {
        extern TProc_Interface* Proc_MusicPlayer_GetInterface(void);
        pProc_Interface=Proc_MusicPlayer_GetInterface();
      } break;
      default: {
        printf("Illigal proc state. ProcStateNext:%d ProcStateLast:%d\n",ProcStateNext,ProcStateLast);
        SystemHalt();
      } break;
    }
    
    PrintFreeMem();
    VRAMManager_ShowFreeMem();
    pProc_Interface->Start(ProcStateLast);
    PrintFreeMem();
    VRAMManager_ShowFreeMem();
    ProcStateLast=ProcStateNext;
    ProcStateNext=EPS_Loop;
    
    VRAMManager_VariableMode_Start();
    
    VBlankPassedCount=0;
    
    while(1){
      if(SystemExitRequest==true) break;
      
      u32 VSyncCount=VBlankPassedCount;
      VBlankPassedCount=0;
      if(4<VSyncCount) VSyncCount=4;
      
      main_ProcessKeys(VSyncCount);
      
      GuStart();
      pProc_Interface->UpdateGU(VSyncCount);
      GuFinish();
      
      if(VBlankPassed==false) sceDisplayWaitVblankStart();
      VBlankPassed=false;
      
      sceGuSync(0,0);
      GuSwapBuffers();
      
      PlayList_Update();
      
      if(ProcStateNext!=EPS_Loop) break;
    }
    
    VRAMManager_VariableMode_End();
    
    PrintFreeMem();
    VRAMManager_ShowFreeMem();
    pProc_Interface->End();
    PrintFreeMem();
    VRAMManager_ShowFreeMem();
    
    KeyRepeat_Free();
    
    VRAMManager_ProcessMode_End();
    
    if(SystemExitRequest==true) break;
  }
}

int main(void)
{
//  pspDebugScreenInit();
//  pspDebugScreenSetXY(0,0);
  
//  for(u32 idx=0;idx<60;idx++) sceDisplayWaitVblankStart();
  printf("\n");
  
  ChangeSystemFreq_High();
  
  main_Start();
  
  main_Loop();
  
  main_End();
  
  ChangeSystemFreq_Normal();
  
  printf("Terminated. Return to home.\n");
  
  sceKernelExitGame();
  
  return 0;
}

